from dataclasses import dataclass

from marshmallow import Schema, fields, post_load


@dataclass()
class Client:
    userId: str
    robotId: str
    sessionId: str

    def __repr__(self):
        return f"<Client(userId={self.userId}, robotId={self.robotId}, sessionId={self.sessionId})>"  # noqa: E501, B950

    def dump(self):
        return ClientSchema().dump(self)

    def dumps(self):
        return ClientSchema().dumps(self)

    @staticmethod
    def load(data):
        return ClientSchema().load(data)


class ClientSchema(Schema):
    userId = fields.String(required=True)
    robotId = fields.String(required=True)
    sessionId = fields.String(required=True)

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return Client(**data)



# This will become common across all SDKs so we can't assume it's going to flake cleanly.
# flake8: noqa

# Very generic find query
GQL_FRAGMENT_VARIABLES = """
fragment ppe_fields on PPE {
  solveKey
  suggested
  max
  mean
  lastUpdatedTimestamp {formatted}
}
fragment solverdata_fields on SOLVERDATA {
  solveKey
  BayesNetOutVertIDs
  BayesNetVertID
  dimIDs
  dimbw
  dims
  dimval
  dontmargin
  eliminated
  infoPerCoord
  initialized
  ismargin
  separator
  solveInProgress
  solvedCount
  variableType
  vecbw
  vecval
  _version
}
fragment variable_skeleton_fields on VARIABLE {
	label
  tags
}
fragment variable_summary_fields on VARIABLE {
  timestamp {formatted}
  ppes {
    ...ppe_fields
  }
  variableType
  _version
  _id
}
fragment variable_full_fields on VARIABLE{
  smallData
  solvable
  solverData
  {
		...solverdata_fields
  }
}
"""

GQL_FRAGMENT_FACTORS = """
fragment factor_skeleton_fields on FACTOR {
	label
  tags
  _variableOrderSymbols
}
fragment factor_summary_fields on FACTOR {
  timestamp {formatted}
  _version
}
fragment factor_full_fields on FACTOR {
  fnctype
  solvable
  data
}
"""

GQL_GETVARIABLES = """
query sdk_get_variables(
  	$userId: ID!,
  	$robotIds: [ID!]!,
  	$sessionIds: [ID!]!,
    $variable_label_regexp: ID = ".*",
    $variable_tags: [String!] = ["VARIABLE"],
    $solvable: Int! = 0,
  	$fields_summary: Boolean! = false,
  	$fields_full: Boolean! = false){
	USER(id: $userId) {
    name
		robots(filter:{id_in: $robotIds}) {
      name
      sessions(filter:{id_in: $sessionIds}){
        name
        variables(filter:{
            label_regexp: $variable_label_regexp,
          	tags_contains: $variable_tags,
          	solvable_gte: $solvable}) {
          ...variable_skeleton_fields # Always include
          ...variable_summary_fields @include(if: $fields_summary)
          ...variable_full_fields @include(if: $fields_full)
        }
      }
    }
  }
}"""

GQL_GETFACTORS = """
query sdk_get_factors(
  	$userId: ID!,
  	$robotIds: [ID!]!,
  	$sessionIds: [ID!]!,
    $factor_label_regexp: ID = ".*",
    $factor_tags: [String!] = ["FACTOR"],
    $solvable: Int! = 0,
  	$fields_summary: Boolean! = false,
  	$fields_full: Boolean! = false){
	USER(id: $userId) {
    name
		robots(filter:{id_in: $robotIds}) {
      name
      sessions(filter:{id_in: $sessionIds}){
        name
        factors(filter:{
            label_regexp: $factor_label_regexp,
          	tags_contains: $factor_tags,
          	solvable_gte: $solvable}) {
          ...factor_skeleton_fields # Always include
          ...factor_summary_fields @include(if: $fields_summary)
          ...factor_full_fields @include(if: $fields_full)
        }
      }
    }
  }
}"""


GQL_GETVARIABLESFACTORS = """
query sdk_get_variablesfactors(
  	$userId: ID!,
  	$robotIds: [ID!]!,
  	$sessionIds: [ID!]!,
  	$variables: Boolean! = true,
  	$factors: Boolean! = true,
    $variable_label_regexp: ID = ".*",
    $factor_label_regexp: ID = ".*",
    $variable_tags: [String!] = ["VARIABLE"],
    $factor_tags: [String!] = ["FACTOR"],
    $solvable: Int! = 0,
  	$fields_summary: Boolean! = false,
  	$fields_full: Boolean! = false){
	USER(id: $userId) {
    name
		robots(filter:{id_in: $robotIds}) {
      name
      sessions(filter:{id_in: $sessionIds}){
        name
        variables(filter:{
            label_regexp: $variable_label_regexp,
          	tags_contains: $variable_tags,
          	solvable_gte: 0}) @include(if: $variables){
          ...variable_skeleton_fields # Always include
          ...variable_summary_fields @include(if: $fields_summary)
          ...variable_full_fields @include(if: $fields_full)
        }
        factors(filter:{
            label_regexp: $factor_label_regexp,
          	tags_contains: $factor_tags,
          	solvable_gte: $solvable}) @include(if: $factors){
          ...factor_skeleton_fields # Always include
          ...factor_summary_fields @include(if: $fields_summary)
          ...factor_full_fields @include(if: $fields_full)
        }
      }
    }
  }
}"""


GQL_GETVARIABLE = """
query sdk_get_variable(
  	$userId: ID!,
  	$robotId: ID!,
  	$sessionId: ID!,
    $label: ID!) {
	USER(id: $userId) {
		robots(filter:{id: $robotId}) {
      sessions(filter:{id: $sessionId}) {
        variables(filter:{label: $label}) {
          ...variable_skeleton_fields
          ...variable_summary_fields
          ...variable_full_fields
        }
      }
    }
  }
}"""


GQL_GETFACTOR = """
query sdk_get_variable(
  	$userId: ID!,
  	$robotId: ID!,
  	$sessionId: ID!,
    $label: ID!) {
	USER(id: $userId) {
		robots(filter:{id: $robotId}) {
      sessions(filter:{id: $sessionId}) {
        factors(filter:{label: $label}) {
          ...factor_skeleton_fields
          ...factor_summary_fields
          ...factor_full_fields
        }
      }
    }
  }
}"""


GQL_GETSTATUSMESSAGES = """
query sdk_ls_statusmessages($id: ID!) {
    statusMessages(id: $id) {
        requestId,
        action,
        state,
        timestamp,
        client {
            userId,
            robotId,
            sessionId
        }
    }
}
"""

GQL_GETSTATUSLATEST = """
query sdk_get_statuslatest($id: ID!) {
  statusLatest(id: $id) {
    requestId,
    action,
    state,
    timestamp,
    client {
      userId,
      robotId,
      sessionId
    }
  }
}
"""



from dataclasses import dataclass, field
from typing import List

import numpy
from marshmallow import EXCLUDE, Schema, fields, post_load

from navability.common.versions import payload_version


@dataclass()
class VariableNodeData:
    variableType: str
    vecval: List[float] = field(default_factory=lambda: list(numpy.zeros(3 * 100)))
    dimval: int = 3
    vecbw: List[float] = field(default_factory=lambda: list(numpy.zeros(3)))
    dimbw: int = 3
    BayesNetOutVertIDs: List[int] = field(default_factory=list)
    dimIDs: List[int] = field(default_factory=lambda: [0, 1, 2])
    dims: int = 3
    eliminated: bool = False
    BayesNetVertID: str = "_null"
    separator: List[int] = field(default_factory=list)
    initialized: bool = False
    infoPerCoord: List[int] = field(default_factory=lambda: list(numpy.zeros(3)))
    ismargin: bool = False
    dontmargin: bool = False
    solveInProgress: int = 0
    solvedCount: int = 0
    solveKey: str = "default"
    _version: str = payload_version

    def __repr__(self):
        return f"<VariableNodeData(solveKey={self.solveKey})>"

    def dump(self):
        return VariableNodeDataSchema().dump(self)

    def dumps(self):
        return VariableNodeDataSchema().dumps(self)

    @staticmethod
    def load(data):
        return VariableNodeDataSchema().load(data)


class VariableNodeDataSchema(Schema):
    vecval = fields.List(fields.Float(), required=True)  # numpy.zeros(3*100) # 300
    dimval = fields.Integer(required=True)  # 3
    vecbw = fields.List(fields.Float(), required=True)  # numpy.zeros(3)
    dimbw = fields.Integer(required=True)  # 3
    BayesNetOutVertIDs = fields.List(fields.Integer(), required=True)  # []
    dimIDs = fields.List(fields.Integer(), required=True)  # [0,1,2]
    dims = fields.Integer(required=True)  # 3
    eliminated = fields.Boolean(required=True)  # False
    BayesNetVertID = fields.Str(required=True)  # "_null"
    separator = fields.List(fields.Integer(), required=True)  # []
    variableType = fields.Str(required=True)  # type
    initialized = fields.Boolean(required=True)  # False
    infoPerCoord = fields.List(fields.Float(), required=True)  # numpy.zeros(3)
    ismargin = fields.Boolean(required=True)  # False
    dontmargin = fields.Boolean(required=True)  # False
    solveInProgress = fields.Integer(required=True)  # 0
    solvedCount = fields.Integer(required=True)  # 0
    solveKey = fields.Str(required=True)  # solveKey
    _version: fields.Str(data_key="_version", required=False)

    class Meta:
        ordered = True
        unknown = EXCLUDE  # Note: This is because of _version, remote and fix later.

    @post_load
    def marshal(self, data, **kwargs):
        return VariableNodeData(**data)



from dataclasses import dataclass, field

import numpy


class Distribution:
    """
    Base type for all distribution classes.
    """

    pass

    def dumpsPacked(self):
        raise Exception(f"'dumpsPacked' has not been implemented in {str(type(self))}")


@dataclass
class FullNormal(Distribution):
    mean: numpy.ndarray = field(default_factory=lambda: numpy.zeros(3))
    covariance: numpy.ndarray = field(
        default_factory=lambda: numpy.diag([0.1, 0.1, 0.1])
    )

    def dumpsPacked(self):
        means = (
            str(self.mean.tolist())
            if isinstance(self.mean, numpy.ndarray)
            else str(self.mean)
        )
        covars = ";".join(
            [
                str(v.tolist()).replace("[", "").replace("]", "").replace(",", "")
                for v in self.covariance
            ]
        )
        return (
            f"FullNormal(\ndim: {len(self.mean)}\nμ: {means}\nΣ: [{covars}]\n)\n"
        )  # noqa: E501, B950

    @staticmethod
    def load(data: str):
        # REF: IncrementalInference/src/SerializingDistributions.jl:111
        mean = numpy.asarray(
            [
                float(v)
                for v in data.split("μ")[1].split("]")[0].split("[")[1].split(",")
            ]
        )
        rows = data.split("Σ")[1].split("]")[0].split("[")[1].split(";")
        covariance = numpy.asmatrix(
            [[float(v) for v in r.split(" ") if v != ""] for r in rows]
        )
        return FullNormal(mean=mean, covariance=covariance)

    def __eq__(self, __o: object) -> bool:
        return (self.mean == __o.mean).all() and (
            self.covariance == __o.covariance
        ).all()



from dataclasses import dataclass, field
from typing import List

from marshmallow import Schema, fields, post_load


@dataclass()
class Scope:
    environmentIds: List[str] = field(default_factory=lambda: [])
    userIds: List[str] = field(default_factory=lambda: [])
    robotIds: List[str] = field(default_factory=lambda: [])
    sessionIds: List[str] = field(default_factory=lambda: [])

    def dump(self):
        return ScopeSchema().dump(self)

    def dumps(self):
        return ScopeSchema().dumps(self)

    @staticmethod
    def load(data: str):
        return ScopeSchema().load(data)


class ScopeSchema(Schema):
    environmentIds = fields.String(many=True, default=[])
    userIds = fields.String(many=True, default=[])
    robotIds = fields.String(many=True, default=[])
    sessionIds = fields.String(many=True, default=[])

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return Scope(**data)



import json
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from marshmallow import EXCLUDE, Schema, fields, post_load

from navability.common.timestamps import TS_FORMAT
from navability.common.versions import payload_version


@dataclass()
class FactorSkeleton:
    label: str
    variableOrderSymbols: List[str]
    tags: List[str]

    def __repr__(self):
        return (
            f"<FactorSkeleton(label={self.label},"
            f"variableOrderSymbols={self.variableOrderSymbols},tags={self.tags})>"
        )

    def dump(self):
        return FactorSkeletonSchema().dump(self)

    def dumps(self):
        return FactorSkeletonSchema().dumps(self)

    @staticmethod
    def load(data):
        return FactorSkeletonSchema().load(data)


class FactorSkeletonSchema(Schema):
    label = fields.Str(required=True)
    variableOrderSymbols = fields.List(
        fields.Str(), data_key="_variableOrderSymbols", required=True
    )
    tags = fields.List(fields.Str(), required=True)

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return FactorSkeleton(**data)


@dataclass()
class FactorSummary:
    label = str
    variableOrderSymbols: List[str]
    tags: List[str] = field(default_factory=lambda: ["FACTOR"])
    timestamp: datetime = datetime.utcnow()
    _version: str = payload_version

    def __repr__(self):
        return (
            f"<FactorSkeleton(label={self.label},"
            f"variableOrderSymbols={self.variableOrderSymbols},tags={self.tags})>"
        )

    def dump(self):
        return FactorSummarySchema().dump(self)

    def dumps(self):
        return FactorSummarySchema().dumps(self)

    @staticmethod
    def load(data):
        return FactorSummarySchema().load(data)


class FactorSummarySchema(Schema):
    label = fields.Str(required=True)
    variableOrderSymbols: fields.Str(data_key="_variableOrderSymbols", many=True)
    tags = fields.List(fields.Str(), required=True)
    timestamp = fields.Method("get_timestamp", "set_timestamp", required=True)
    _version = fields.Str(required=True)

    class Meta:
        ordered = True
        unknown = EXCLUDE  # Note: This is because of _version, remote and fix later.

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.timestamp.isoformat(timespec="milliseconds")
        if not obj.timestamp.tzinfo:
            ts += "Z"
        return ts

    def set_timestamp(self, obj):
        # Have to be defensive here because it could be simply serialized
        # or it can be GQL data with formatted
        tsraw = obj if type(obj) == str else obj["formatted"]
        return datetime.strptime(tsraw, TS_FORMAT)

    @post_load
    def marshal(self, data, **kwargs):
        return FactorSummary(**data)


@dataclass()
class FactorData:
    fnc: Dict  # InferenceType  # {"datastr":"FullNormal(\\ndim: 3\\nμ: [10.0, 0.0, 1.0471975511965976]\\nΣ: [0.010000000000000002 0.0 0.0; 0.0 0.010000000000000002 0.0; 0.0 0.0 0.010000000000000002]\\n)\\n"}  # noqa: E501, B950
    eliminated: bool = False
    potentialused: bool = False
    edgeIDs: List[int] = field(default_factory=lambda: [])
    multihypo: List[int] = field(default_factory=lambda: [])
    certainhypo: List[int] = field(default_factory=lambda: [1, 2])
    nullhypo: float = 0
    solveInProgress: int = 0
    inflation: float = 5

    def dumps(self):
        return FactorDataSchema().dumps(self)

    def dump(self):
        return FactorDataSchema().dump(self)


class FactorDataSchema(Schema):
    eliminated = fields.Bool(required=True)
    potentialused = fields.Bool(required=True)
    edgeIDs = fields.List(fields.Int(), required=True)
    fnc = fields.Dict(
        required=True
    )  # fields.Method("get_fnc", "set_fnc", required=True)
    multihypo = fields.List(fields.Int(), required=True)
    certainhypo = fields.List(fields.Int(), required=True)
    nullhypo = fields.Float(required=True)
    solveInProgress = fields.Int(required=True)
    inflation = fields.Float(required=True)

    class Meta:
        ordered = True

    # def get_fnc(self, obj):
    #     # TODO: Switch this out to a real embedded object, no need for strings.
    #     return obj.fnc.dump()

    # def set_fnc(self, ob):
    #     raise Exception("Deserialization not supported yet.")

    @post_load
    def marshal(self, data, **kwargs):
        return FactorData(**data)


@dataclass()
class Factor:
    label: str
    fnctype: str
    variableOrderSymbols: List[str]
    data: FactorData  # '{"eliminated":false,"potentialused":false,"edgeIDs":[],"fnc":{"datastr":"FullNormal(\\ndim: 3\\nμ: [10.0, 0.0, 1.0471975511965976]\\nΣ: [0.010000000000000002 0.0 0.0; 0.0 0.010000000000000002 0.0; 0.0 0.0 0.010000000000000002]\\n)\\n"},"multihypo":[],"certainhypo":[1,2],"nullhypo":0.0,"solveInProgress":0,"inflation":5.0}'  # noqa: E501, B950
    tags: List[str] = field(default_factory=lambda: ["FACTOR"])
    timestamp: str = datetime.utcnow()
    nstime: int = 0
    solvable: str = 1
    _version: str = payload_version

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}"
            f"(label={self.label},"
            f"variables={self.variableOrderSymbols})>"
        )

    def dump(self):
        return FactorSchema().dump(self)

    def dumps(self):
        return FactorSchema().dumps(self)

    @staticmethod
    def load(data):
        return FactorSchema().load(data)


class FactorSchema(Schema):
    label = fields.Str(required=True)
    _version = fields.Str(required=True)
    _variableOrderSymbols = fields.Method("get_variableOrderSymbols", required=True)
    data = fields.Method("get_data", "set_data", required=True)
    tags = fields.List(fields.Str(), required=True)
    timestamp = fields.Method("get_timestamp", "set_timestamp", required=True)
    nstime = fields.Str(default="0")
    fnctype = fields.Str(required=True)
    solvable = fields.Int(required=True)

    class Meta:
        ordered = True
        unknown = EXCLUDE  # Note: This is because of _version, remote and fix later.

    def get_variableOrderSymbols(self, obj):
        # TODO: Switch this out to a real embedded object, no need for strings.
        return json.dumps(obj.variableOrderSymbols)

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.timestamp.isoformat(timespec="milliseconds")
        if not obj.timestamp.tzinfo:
            ts += "Z"
        return ts

    def set_timestamp(self, obj):
        # Have to be defensive here because it could be simply serialized
        # or it can be GQL data with formatted
        tsraw = obj if type(obj) == str else obj["formatted"]
        return datetime.strptime(tsraw, TS_FORMAT)

    def get_data(self, obj):
        return obj.data.dumps()

    def set_data(self, ob):
        raise Exception("Deserialization not supported yet.")

    @post_load
    def marshal(self, data, **kwargs):
        return Factor(**data)



from dataclasses import dataclass

from gql import Client as GQLCLient
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport


@dataclass(frozen=True)
class QueryOptions:
    query: str
    variables: any = None
    fetchPolicy: any = None


@dataclass(frozen=True)
class MutationOptions:
    mutation: str
    variables: any = None
    fetchPolicy: any = None


class NavAbilityClient:
    def query(self, options: QueryOptions):
        pass

    def mutate(self, options: MutationOptions):
        pass


class NavAbilityWebsocketClient(NavAbilityClient):
    def __init__(self, url: str = "wss://api.d1.navability.io/graphql") -> None:
        super().__init__()
        transport = WebsocketsTransport(url=url)
        self.client = GQLCLient(transport=transport, fetch_schema_from_transport=True)

    def query(self, options: QueryOptions):
        return self.client.execute(options.query, options.variables)

    def mutate(self, options: MutationOptions):
        return self.client.execute(options.mutation, options.variables)


class NavAbilityHttpsClient(NavAbilityClient):
    def __init__(self, url: str = "https://api.d1.navability.io") -> None:
        super().__init__()
        transport = AIOHTTPTransport(url=url)
        self.client = GQLCLient(transport=transport, fetch_schema_from_transport=True)

    def query(self, options: QueryOptions):
        return self.client.execute(options.query, options.variables)

    def mutate(self, options: MutationOptions):
        return self.client.execute(options.mutation, options.variables)



# flake8: noqa: F401
from .client import Client
from .factor.distributions import Distribution, FullNormal
from .factor.factor import Factor, FactorData, FactorSkeleton, FactorSummary
from .factor.inferencetypes import (
    InferenceType,
    Pose2AprilTag4Corners,
    Pose2Pose2,
    PriorPose2,
)
from .navabilityclient import (
    NavAbilityClient,
    NavAbilityHttpsClient,
    NavAbilityWebsocketClient,
)
from .querydetail import QueryDetail
from .scope import Scope
from .statusmessage import MutationUpdate, StatusMessage
from .variable.ppe import PPE
from .variable.variable import Variable, VariableSkeleton, VariableSummary, VariableType
from .variable.variablenodedata import VariableNodeData



# This will become common across all SDKs so we can't assume it's going to flake cleanly.
# flake8: noqa

GQL_ADDVARIABLE = """
mutation sdk_addVariable ($variable: FactorGraphInput!) {
    addVariable(variable: $variable)
}
"""

GQL_ADDFACTOR = """
mutation sdk_addFactor ($factor: FactorGraphInput!) {
  addFactor(factor: $factor)
}
"""

GQL_SOLVESESSION = """
mutation sdk_solveSession ($client: ClientInput!) {
  solveSession(client: $client)
}
"""



import logging
from typing import List

from gql import gql

from navability.common.mutations import GQL_ADDVARIABLE
from navability.common.queries import (
    GQL_FRAGMENT_VARIABLES,
    GQL_GETVARIABLE,
    GQL_GETVARIABLES,
)
from navability.entities.client import Client
from navability.entities.navabilityclient import (
    MutationOptions,
    NavAbilityClient,
    QueryOptions,
)
from navability.entities.querydetail import QueryDetail
from navability.entities.variable.variable import (
    Variable,
    VariableSchema,
    VariableSkeleton,
    VariableSkeletonSchema,
    VariableSummarySchema,
)

DETAIL_SCHEMA = {
    QueryDetail.LABEL: None,
    QueryDetail.SKELETON: VariableSkeletonSchema(),
    QueryDetail.SUMMARY: VariableSummarySchema(),
    QueryDetail.FULL: VariableSchema(),
}

logger = logging.getLogger(__name__)


def addVariable(navAbilityClient: NavAbilityClient, client: Client, v: Variable):
    return navAbilityClient.mutate(
        MutationOptions(
            gql(GQL_ADDVARIABLE),
            {"variable": {"client": client.dump(), "packedData": v.dumpsPacked()}},
        )
    )["addVariable"]


def listVariables(
    navAbilityClient: NavAbilityClient,
    client: Client,
    regexFilter: str = ".*",
    tags: List[str] = None,
    solvable: int = 0,
) -> List[str]:
    return [
        v.label
        for v in getVariables(
            navAbilityClient,
            client,
            detail=QueryDetail.SKELETON,
            regexFilter=regexFilter,
            tags=tags,
            solvable=solvable,
        )
    ]


# Alias
ls = listVariables


def getVariables(
    navAbilityClient: NavAbilityClient,
    client: Client,
    detail: QueryDetail = QueryDetail.SKELETON,
    regexFilter: str = ".*",
    tags: List[str] = None,
    solvable: int = 0,
) -> List[VariableSkeleton]:
    params = {
        "userId": client.userId,
        "robotIds": [client.robotId],
        "sessionIds": [client.sessionId],
        "variable_label_regexp": regexFilter,
        "variable_tags": tags if tags is not None else ["VARIABLE"],
        "solvable": solvable,
        "fields_summary": detail in [QueryDetail.SUMMARY, QueryDetail.FULL],
        "fields_full": detail == QueryDetail.FULL,
    }
    logger.debug(f"Query params: {params}")
    res = navAbilityClient.query(
        QueryOptions(gql(GQL_FRAGMENT_VARIABLES + GQL_GETVARIABLES), params)
    )
    logger.debug(f"Query result: {res}")
    # TODO: Check for errors
    schema = DETAIL_SCHEMA[detail]
    # Using the hierarchy approach, we need to check that we have
    # exactly one user/robot/session in it, otherwise error.
    if (
        "USER" not in res
        or len(res["USER"]) != 1
        or len(res["USER"][0]["robots"]) != 1
        or len(res["USER"][0]["robots"][0]["sessions"]) != 1
        or "variables" not in res["USER"][0]["robots"][0]["sessions"][0]
    ):
        # Debugging information
        if len(res["USER"]) != 1:
            logger.warn("User not found in result, returning empty list")
        if len(res["USER"][0]["robots"]) != 1:
            logger.warn("Robot not found in result, returning empty list")
        if len(res["USER"][0]["robots"][0]["sessions"]) != 1:
            logger.warn("Robot not found in result, returning empty list")
        return []
    if schema is None:
        return res["USER"][0]["robots"][0]["sessions"][0]["variables"]
    return [
        schema.load(l) for l in res["USER"][0]["robots"][0]["sessions"][0]["variables"]
    ]


def getVariable(navAbilityClient: NavAbilityClient, client: Client, label: str):
    params = client.dump()
    params["label"] = label
    logger.debug(f"Query params: {params}")
    res = navAbilityClient.query(
        QueryOptions(gql(GQL_FRAGMENT_VARIABLES + GQL_GETVARIABLE), params)
    )
    logger.debug(f"Query result: {res}")
    # TODO: Check for errors
    # Using the hierarchy approach, we need to check that we have
    # exactly one user/robot/session in it, otherwise error.
    if (
        "USER" not in res
        or len(res["USER"][0]["robots"]) != 1
        or len(res["USER"][0]["robots"][0]["sessions"]) != 1
        or "variables" not in res["USER"][0]["robots"][0]["sessions"][0]
    ):
        raise Exception(
            "Received an empty data structure, set logger to debug for the payload"
        )
    vs = res["USER"][0]["robots"][0]["sessions"][0]["variables"]
    # TODO: Check for errors
    if len(vs) == 0:
        return None
    if len(vs) > 1:
        raise Exception(f"More than one variable named {label} returned")
    return Variable.load(vs[0])



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import List

from marshmallow import Schema, fields, post_load

from navability.common.timestamps import TS_FORMAT, TS_FORMAT_NO_TZ


@dataclass()
class PPE:
    solveKey: str
    suggested: List[float]
    max: List[float]
    mean: List[float]
    lastUpdatedTimestamp: datetime

    def __repr__(self):
        return (
            f"<PPE(solveKey={self.solveKey},suggested={self.suggested},"
            f"lastUpdatedTimestamp={self.lastUpdatedTimestamp})>"
        )

    def dump(self):
        return PPE.schema.dump(self)

    def dumps(self):
        return PPE.schema.dumps(self)

    @staticmethod
    def load(data):
        return PPESchema().load(data)


class PPESchema(Schema):
    solveKey = fields.Str(required=True)
    suggested = fields.List(fields.Float(), required=True)
    max = fields.List(fields.Float(), required=True)
    mean = fields.List(fields.Float(), required=True)
    lastUpdatedTimestamp = fields.Method(
        "get_timestamp", "set_timestamp", required=True
    )

    class Meta:
        ordered = True

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.lastUpdatedTimestamp.isoformat(timespec="milliseconds")
        if not obj.lastUpdatedTimestamp.tzinfo:
            ts += "Z"
        return ts

    def set_timestamp(self, obj):
        # Have to be defensive here because it could be simply serialized
        # or it can be GQL data with formatted
        tsraw = obj if type(obj) == str else obj["formatted"]
        # Defensively append the Z if needed
        try:
            ts = datetime.strptime(tsraw, TS_FORMAT)
        except (ValueError):
            # Fall back with no timezone and assign it.
            # NOTE: This happens with PPEs, we need to update solver PPEs it tz.
            ts = datetime.strptime(tsraw, TS_FORMAT_NO_TZ).replace(tzinfo=timezone.utc)
        return ts

    @post_load
    def marshal(self, data, **kwargs):
        return PPE(**data)



from dataclasses import dataclass

import numpy as np
from marshmallow import Schema, fields, post_load

from navability.entities.factor.distributions import Distribution


@dataclass
class InferenceType:
    """
    Base type for all factor classes.
    """

    pass


@dataclass
class PriorPose2(InferenceType):
    z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(z={str(self.z)})>"

    def dump(self):
        return PriorPose2Schema().dump(self)

    def dumps(self):
        return PriorPose2Schema().dumps(self)

    # TODO: Deserializing this.


class PriorPose2Schema(Schema):
    z = fields.Method("get_packed", "set_packed", data_key="str", required=True)

    class Meta:
        ordered = True

    def get_packed(self, obj):
        return obj.z.dumpsPacked()

    def set_packed(self, obj):
        raise Exception("This has not been implemented yet.")

    # @post_load
    # def load(self, data, **kwargs):
    #     return PriorPose2(**data)


@dataclass
class Pose2Pose2(InferenceType):
    z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(z={str(self.z)})>"

    def dump(self):
        return Pose2Pose2Schema().dump(self)

    def dumps(self):
        return Pose2Pose2Schema().dumps(self)

    # TODO: Deserializing this.


class Pose2Pose2Schema(Schema):
    datastr = fields.Method("get_packed", "set_packed", required=True)

    def get_packed(self, obj):
        return obj.z.dumpsPacked()

    def set_packed(self, obj):
        raise Exception("This has not been implemented yet.")

    @post_load
    def marshal(self, data, **kwargs):
        return PriorPose2(**data)


@dataclass
class Pose2AprilTag4Corners(InferenceType):
    corners: np.ndarray
    homography: np.ndarray
    K: np.ndarray
    taglength: float
    id: int
    mimeType: str = "/application/JuliaLang/PackedPose2AprilTag4Corners"

    def dump(self):
        return Pose2AprilTag4CornersSchema().dump(self)

    def dumps(self):
        return Pose2AprilTag4CornersSchema().dumps(self)

    # TODO: Deserializing this.


class Pose2AprilTag4CornersSchema(Schema):
    corners = fields.List(fields.Float, required=True)
    homography = fields.List(fields.Float, required=True)
    K = fields.List(fields.Float, required=True)
    taglength = fields.Float(required=True)
    id = fields.Int(required=True)
    mimeType = fields.String(
        default="/application/JuliaLang/PackedPose2AprilTag4Corners"
    )

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return Pose2AprilTag4Corners(**data)



from gql import gql

from navability.common.mutations import GQL_SOLVESESSION
from navability.entities.client import Client
from navability.entities.navabilityclient import MutationOptions, NavAbilityClient


def solveSession(navAbilityClient: NavAbilityClient, client: Client):
    return navAbilityClient.mutate(
        MutationOptions(gql(GQL_SOLVESESSION), {"client": client.dump()})
    )["solveSession"]



from dataclasses import dataclass

from marshmallow import Schema, fields, post_load

from navability.entities.client import Client, ClientSchema


@dataclass()
class StatusMessage:
    requestId: str
    action: str
    state: str
    timestamp: str
    client: Client

    def __repr__(self):
        return (
            f"<StatusMessage(requestId={self.requestId}, "
            f"timestamp={self.timestamp}, client={self.client}, "
            f"action={self.action}, state={self.state})>"
        )

    def dump(self):
        return StatusMessage().dump(self)

    def dumps(self):
        return StatusMessage().dumps(self)

    @staticmethod
    def load(data):
        return StatusMessageSchema().load(data)


class StatusMessageSchema(Schema):
    requestId = fields.String(required=True)
    action = fields.String(required=True)
    state = fields.String(required=True)
    timestamp = fields.DateTime(required=True)
    client = fields.Nested(ClientSchema, required=True)

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return StatusMessage(**data)


@dataclass
class MutationUpdate:
    mutationUpdate: StatusMessage

    def dump(self):
        return MutationUpdate().dump(self)

    def dumps(self):
        return MutationUpdate().dumps(self)

    @staticmethod
    def load(data):
        return MutationUpdate().load(data)


class MutationUpdateSchema(Schema):
    mutationUpdate = fields.Nested(StatusMessageSchema, required=True)

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return MutationUpdate(**data)



import logging
from typing import List

from gql import gql

from navability.common.mutations import GQL_ADDFACTOR
from navability.common.queries import (
    GQL_FRAGMENT_FACTORS,
    GQL_GETFACTOR,
    GQL_GETFACTORS,
)
from navability.entities.client import Client
from navability.entities.factor.factor import (
    Factor,
    FactorSchema,
    FactorSkeleton,
    FactorSkeletonSchema,
    FactorSummarySchema,
)
from navability.entities.navabilityclient import (
    MutationOptions,
    NavAbilityClient,
    QueryOptions,
)
from navability.entities.querydetail import QueryDetail

DETAIL_SCHEMA = {
    QueryDetail.LABEL: None,
    QueryDetail.SKELETON: FactorSkeletonSchema(),
    QueryDetail.SUMMARY: FactorSummarySchema(),
    QueryDetail.FULL: FactorSchema(),
}

logger = logging.getLogger(__name__)


def addFactor(navAbilityClient: NavAbilityClient, client: Client, f: Factor):
    return navAbilityClient.mutate(
        MutationOptions(
            gql(GQL_ADDFACTOR),
            {"factor": {"client": client.dump(), "packedData": f.dumps()}},
        )
    )["addFactor"]


def listFactors(
    navAbilityClient: NavAbilityClient,
    client: Client,
    regexFilter: str = ".*",
    tags: List[str] = None,
    solvable: int = 0,
) -> List[str]:
    return [
        v.label
        for v in getFactors(
            navAbilityClient,
            client,
            detail=QueryDetail.SKELETON,
            regexFilter=regexFilter,
            tags=tags,
            solvable=solvable,
        )
    ]


# Alias
lsf = listFactors


def getFactors(
    navAbilityClient: NavAbilityClient,
    client: Client,
    detail: QueryDetail = QueryDetail.SKELETON,
    regexFilter: str = ".*",
    tags: List[str] = None,
    solvable: int = 0,
) -> List[FactorSkeleton]:
    params = {
        "userId": client.userId,
        "robotIds": [client.robotId],
        "sessionIds": [client.sessionId],
        "factor_label_regexp": regexFilter,
        "factor_tags": tags if tags is not None else ["FACTOR"],
        "solvable": solvable,
        "fields_summary": detail in [QueryDetail.SUMMARY, QueryDetail.FULL],
        "fields_full": detail == QueryDetail.FULL,
    }
    logger.debug(f"Query params: {params}")
    res = navAbilityClient.query(
        QueryOptions(gql(GQL_FRAGMENT_FACTORS + GQL_GETFACTORS), params)
    )
    logger.debug(f"Query result: {res}")
    # TODO: Check for errors
    schema = DETAIL_SCHEMA[detail]
    # Using the hierarchy approach, we need to check that
    # we have exactly one user/robot/session in it, otherwise error.
    if (
        "USER" not in res
        or len(res["USER"]) != 1
        or len(res["USER"][0]["robots"]) != 1
        or len(res["USER"][0]["robots"][0]["sessions"]) != 1
        or "factors" not in res["USER"][0]["robots"][0]["sessions"][0]
    ):
        # Debugging information
        if len(res["USER"]) != 1:
            logger.warn("User not found in result, returning empty list")
        if len(res["USER"][0]["robots"]) != 1:
            logger.warn("Robot not found in result, returning empty list")
        if len(res["USER"][0]["robots"][0]["sessions"]) != 1:
            logger.warn("Robot not found in result, returning empty list")
        return []
    if schema is None:
        return res["USER"][0]["robots"][0]["sessions"][0]["factors"]
    return [
        schema.load(l) for l in res["USER"][0]["robots"][0]["sessions"][0]["factors"]
    ]


def getFactor(navAbilityClient: NavAbilityClient, client: Client, label: str):
    params = client.dump()
    params["label"] = label
    logger.debug(f"Query params: {params}")
    res = navAbilityClient.query(
        QueryOptions(gql(GQL_FRAGMENT_FACTORS + GQL_GETFACTOR), params)
    )
    logger.debug(f"Query result: {res}")
    # TODO: Check for errors
    # Using the hierarchy approach, we need to check that we
    # have exactly one user/robot/session in it, otherwise error.
    if (
        "USER" not in res
        or len(res["USER"][0]["robots"]) != 1
        or len(res["USER"][0]["robots"][0]["sessions"]) != 1
        or "variables" not in res["USER"][0]["robots"][0]["sessions"][0]
    ):
        raise Exception(
            "Received an empty data structure, set logger to debug for the payload"
        )
    fs = res["USER"][0]["robots"][0]["sessions"][0]["factors"]
    # TODO: Check for errors
    if len(fs) == 0:
        return None
    if len(fs) > 1:
        raise Exception(f"More than one factor named {label} returned")
    return Factor.load(fs[0])



from gql import gql

from navability.common.queries import GQL_GETSTATUSLATEST, GQL_GETSTATUSMESSAGES
from navability.entities.navabilityclient import NavAbilityClient, QueryOptions
from navability.entities.statusmessage import StatusMessageSchema


def getStatusMessages(navAbilityClient: NavAbilityClient, id: str):
    statusMessages = navAbilityClient.query(
        QueryOptions(gql(GQL_GETSTATUSMESSAGES), {"id": id})
    )
    schema = StatusMessageSchema(many=True)
    return schema.load(statusMessages["statusMessages"])


def getStatusLatest(navAbilityClient: NavAbilityClient, id: str):
    statusMessages = navAbilityClient.query(
        QueryOptions(gql(GQL_GETSTATUSLATEST), {"id": id})
    )
    schema = StatusMessageSchema()
    return schema.load(statusMessages["statusLatest"])



import time
from typing import List

from navability.entities import NavAbilityClient
from navability.services import getStatusLatest


def waitForCompletion(
    navAbilityClient: NavAbilityClient,
    requestIds: List[str],
    maxSeconds: int = 60,
    expectedStatus: str = "Complete",
    exceptionMessage: str = "Requests did not complete in time",
):
    """Wait for the requests to complete, poll until done.

    Args:
        requestIds (List[str]): The request IDs that should be polled.
        maxSeconds (int, optional): Maximum wait time. Defaults to 60.
        expectedStatus (str, optional): Expected status message per request.
            Defaults to "Complete".
    """
    wait_time = maxSeconds
    while any(
        [
            getStatusLatest(navAbilityClient, res).state != expectedStatus
            for res in requestIds
        ]
    ):
        time.sleep(2)
        wait_time -= 2
        if wait_time <= 0:
            raise Exception(exceptionMessage)



import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 0):
    sys.exit(
        """
######################################
# Python 3 is needed #
######################################
"""
    )

_version = "0.1.3"

setup(
    name="navabilitysdk",
    version=_version,
    license="Apache-2.0",
    author="NavAbility",
    author_email="info@navability.io",
    package_dir={"": "src"},
    include_package_data=True,
    packages=find_packages("src", exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    # entry_points={"console_scripts": ["navability = navability.main:cli"]},
    python_requires=">=3.8",
    download_url=f"https://github.com/NavAbility/NavAbilitySDK.py/archive/refs/tags/v{_version}.tar.gz",  # noqa: E501, B950
    long_description="""NavAbility SDK: Access NavAbility Cloud factor graph features from Python.
Note that this SDK and the related API are still in development. Please let us know if you have any issues at info@navability.io.""",
    install_requires=[
        "click>=8.0.0,<9",
        "gql[all]==3.0.0a6",
        "marshmallow==3.14.0",
        "numpy>=1.21",
        # Dev/test dependencies
        "black==21.9b0",
        "flake8==4.0.1",
        "pytest==6.2.5",
    ],
)



import json
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Dict, List

from marshmallow import EXCLUDE, Schema, fields, post_load

from navability.common.timestamps import TS_FORMAT
from navability.common.versions import payload_version
from navability.entities.variable.ppe import PPE, PPESchema
from navability.entities.variable.variablenodedata import (
    VariableNodeData,
    VariableNodeDataSchema,
)


class VariableType(Enum):
    Point2 = "RoME.Point2"
    Pose2 = "RoME.Pose2"


@dataclass()
class VariableSkeleton:
    label: str
    tags: List[str] = field(default_factory=lambda: ["VARIABLE"])

    def dump(self):
        return VariableSkeletonSchema().dump(self)

    def dumps(self):
        return VariableSkeletonSchema().dumps(self)

    @staticmethod
    def load(data):
        return VariableSkeletonSchema().load(data)


class VariableSkeletonSchema(Schema):
    label = fields.Str(required=True)
    tags = fields.List(fields.Str(), required=True)

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return VariableSkeleton(**data)


@dataclass()
class VariableSummary:
    label: str
    variableType: str
    tags: List[str] = field(default_factory=lambda: ["VARIABLE"])
    ppes: Dict[str, PPE] = field(default_factory=lambda: {})
    timestamp: datetime = datetime.utcnow()
    _version: str = payload_version
    _id: int = None

    def __repr__(self):
        return (
            f"<VariableSummary(label={self.label},"
            f"variableType={self.variableType},tags={self.tags})>"
        )

    def dump(self):
        return VariableSummarySchema().dump(self)

    def dumps(self):
        return VariableSummarySchema().dumps(self)

    @staticmethod
    def load(data):
        return VariableSummarySchema().load(data)


class VariableSummarySchema(Schema):
    label = fields.Str(required=True)
    tags = fields.List(fields.Str())
    ppes = fields.Nested(PPESchema, many=True)
    timestamp = fields.Method("get_timestamp", "set_timestamp", required=True)
    variableType = fields.Str(required=True)
    _version = fields.Str(required=True)
    _id = fields.Integer(data_key="_id", required=False)

    class Meta:
        ordered = True

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.timestamp.isoformat(timespec="milliseconds")
        if not obj.timestamp.tzinfo:
            ts += "+00"
        return ts

    def set_timestamp(self, obj):
        return datetime.strptime(obj["formatted"], TS_FORMAT)

    @post_load
    def marshal(self, data, **kwargs):
        return VariableSummary(**data)


@dataclass()
class Variable:
    label: str
    variableType: str
    tags: List[str] = field(default_factory=lambda: ["VARIABLE"])
    ppes: Dict[str, PPE] = field(default_factory=lambda: {})
    timestamp: datetime = datetime.utcnow()
    nstime: int = 0
    dataEntry: str = "{}"
    dataEntryType: str = "{}"
    solverData: Dict[str, VariableNodeData] = field(default_factory=lambda: {})
    smallData: str = "{}"
    solvable: str = 1
    _version: str = payload_version
    _id: int = None

    def __post_init__(self):
        if self.solverData == {}:
            self.solverData["default"] = VariableNodeData(self.variableType)

    def __repr__(self):
        return (
            f"<Variable(label={self.label},variableType={self.variableType},"
            f"tags={self.tags})>"
        )

    def dump(self):
        return VariableSchema().dump(self)

    def dumpPacked(self):
        return PackedVariableSchema().dump(self)

    def dumps(self):
        return VariableSchema().dumps(self)

    def dumpsPacked(self):
        return PackedVariableSchema().dumps(self)

    @staticmethod
    def load(data):
        return VariableSchema().load(data)


class VariableSchema(Schema):
    label = fields.Str(required=True)
    tags = fields.List(fields.Str(), required=True)
    ppes = fields.Method("get_ppes", "set_ppes")
    timestamp = fields.Method("get_timestamp", "set_timestamp", required=True)
    variableType = fields.Str(required=True)
    _version = fields.Str(required=True)
    _id: fields.Integer(data_key="_id", required=False)
    # dataEntry = fields.Str(required=True)
    # dataEntryType = fields.Str(required=True)
    solverData = fields.Method("get_solverdata", "set_solverdata")
    smallData = fields.Str(required=True)
    solvable = fields.Int(required=True)

    class Meta:
        ordered = True
        unknown = EXCLUDE  # Note: This is because of _id, remote and fix later.

    @post_load
    def marshal(self, data, **kwargs):
        return Variable(**data)

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.timestamp.isoformat(timespec="milliseconds")
        if not obj.timestamp.tzinfo:
            ts += "Z"
        return ts

    def set_timestamp(self, obj):
        # Have to be defensive here because it could be simply serialized
        # or it can be GQL data with formatted
        tsraw = obj if type(obj) == str else obj["formatted"]
        return datetime.strptime(tsraw, TS_FORMAT)

    def get_solverdata(self, obj):
        return [sd.dump() for sd in obj.solverData.values()]

    def set_solverdata(self, obj):
        return {sd["solveKey"]: VariableNodeData.load(sd) for sd in obj}

    def get_ppes(self, obj):
        return [ppe.dump() for ppe in obj.ppes.values()]

    def set_ppes(self, obj):
        return {ppe["solveKey"]: PPESchema().load(ppe) for ppe in obj}


class PackedVariableSchema(Schema):
    """
    A special schema for the addVariable call, which is used to
    form a packed variable struct.
    """

    label = fields.Str(required=True)
    dataEntry = fields.Str(required=True)
    nstime = fields.Str(default="0")
    variableType = fields.Str(required=True)
    dataEntryType = fields.Str(required=True)
    ppeDict = fields.Str(attribute="ppes", required=True)
    solverDataDict = fields.Method("get_solver_data_dict", required=True)
    smallData = fields.Str(required=True)
    solvable = fields.Int(required=True)
    tags = fields.List(fields.Str(), required=True)
    timestamp = fields.Method("get_timestamp", required=True)
    _version = fields.Str(required=True)

    class Meta:
        ordered = True

    def get_solver_data_dict(self, obj):
        # TODO: Switch this out to a real embedded object, no need for strings.
        schema = VariableNodeDataSchema()
        vnds = {
            solverKey: schema.dump(vnd) for solverKey, vnd in obj.solverData.items()
        }
        return json.dumps(vnds)

    def get_timestamp(self, obj):
        # Return a robust timestamp
        ts = obj.timestamp.isoformat(timespec="milliseconds")
        if not obj.timestamp.tzinfo:
            ts += "+00"
        return ts
